from inspect import isclass
from pathlib import Path
import pytest
from typing import Literal

import algosdk.abi

from graviton.abi_strategy import ABIStrategy

from pyteal import abi

from tests.abi_roundtrip import ABIRoundtrip
from tests.compile_asserts import assert_teal_as_expected

GAI_ISSUE_2050 = "https://github.com/algorand/go-algorand-internal/issues/2050"
BAD_TEALS = {
    "()": GAI_ISSUE_2050,
}


GAI_ISSUE_2068 = "https://github.com/algorand/go-algorand-internal/issues/2068"
BAD_TYPES = {
    # Reference Types
    "account": GAI_ISSUE_2068,
    "asset": GAI_ISSUE_2068,
    "application": GAI_ISSUE_2068,
    # Transaction Types
    "txn": GAI_ISSUE_2068,
    "pay": GAI_ISSUE_2068,
    "keyreg": GAI_ISSUE_2068,
    "acfg": GAI_ISSUE_2068,
    "afrz": GAI_ISSUE_2068,
    "axfer": GAI_ISSUE_2068,
    "appl": GAI_ISSUE_2068,
}


class NamedTupleInherit(abi.NamedTuple):
    a: abi.Field[abi.Bool]
    b: abi.Field[abi.Address]
    c: abi.Field[abi.Tuple2[abi.Uint64, abi.Bool]]
    d: abi.Field[abi.StaticArray[abi.Byte, Literal[10]]]
    e: abi.Field[abi.StaticArray[abi.Bool, Literal[4]]]
    f: abi.Field[abi.Uint64]


PATH = Path.cwd() / "tests" / "integration"
FIXTURES = PATH / "teal"
GENERATED = PATH / "generated"
ABI_TYPES = [
    abi.Address,
    abi.Bool,
    abi.Byte,
    (abi.String, 0),
    (abi.String, 1),
    (abi.String, 13),
    abi.Uint8,
    abi.Uint16,
    abi.Uint32,
    abi.Uint64,
    abi.Account,
    abi.Asset,
    abi.Application,
    abi.Transaction,
    abi.PaymentTransaction,
    abi.KeyRegisterTransaction,
    abi.AssetConfigTransaction,
    abi.AssetFreezeTransaction,
    abi.AssetTransferTransaction,
    abi.ApplicationCallTransaction,
    abi.Tuple0,
    abi.Tuple1[abi.Bool],
    abi.Tuple1[abi.Byte],
    abi.Tuple1[abi.Uint8],
    abi.Tuple1[abi.Uint16],
    abi.Tuple1[abi.Uint32],
    abi.Tuple1[abi.Uint64],
    abi.Tuple2[abi.Bool, abi.Byte],
    abi.Tuple3[abi.Bool, abi.Uint64, abi.Uint32],
    abi.Tuple3[abi.Byte, abi.Bool, abi.Uint64],
    abi.Tuple3[abi.Uint8, abi.Byte, abi.Bool],
    abi.Tuple3[abi.Uint16, abi.Uint8, abi.Byte],
    abi.Tuple3[abi.Uint32, abi.Uint16, abi.Uint8],
    abi.Tuple3[abi.Uint64, abi.Uint32, abi.Uint16],
    abi.Tuple4[abi.Bool, abi.Byte, abi.Address, abi.String],
    abi.Tuple5[abi.Bool, abi.Byte, abi.Address, abi.String, abi.Uint64],
    abi.StaticArray[abi.Bool, Literal[1]],
    abi.StaticArray[abi.Bool, Literal[42]],
    abi.StaticArray[abi.Uint64, Literal[1]],
    abi.StaticArray[abi.Uint64, Literal[42]],
    abi.StaticBytes[Literal[16]],
    (abi.DynamicArray[abi.Bool], 0),
    (abi.DynamicArray[abi.Bool], 1),
    (abi.DynamicArray[abi.Bool], 42),
    (abi.DynamicArray[abi.Uint64], 0),
    (abi.DynamicArray[abi.Uint64], 1),
    (abi.DynamicArray[abi.Uint64], 42),
    (abi.DynamicArray[abi.Address], 10),
    (abi.DynamicBytes, 36),
    (abi.DynamicArray[abi.StaticArray[abi.Bool, Literal[3]]], 11),
    abi.StaticArray[abi.Tuple1[abi.Bool], Literal[10]],
    (
        abi.DynamicArray[
            abi.Tuple4[
                abi.StaticArray[abi.Byte, Literal[4]],
                abi.Tuple2[abi.Bool, abi.Bool],
                abi.Uint64,
                abi.Address,
            ]
        ],
        7,
    ),
    (
        abi.DynamicArray[
            abi.Tuple5[
                abi.Bool,
                abi.Byte,
                abi.Address,
                abi.String,
                abi.Tuple4[
                    abi.Address,
                    abi.StaticArray[
                        abi.Tuple5[
                            abi.Uint32,
                            abi.DynamicArray[abi.String],
                            abi.StaticArray[abi.Bool, Literal[2]],
                            abi.Tuple1[abi.Byte],
                            abi.Uint8,
                        ],
                        Literal[2],
                    ],
                    abi.String,
                    abi.DynamicArray[abi.Bool],
                ],
            ]
        ],
        2,
    ),
    NamedTupleInherit,
]


def roundtrip_setup(abi_type):
    dynamic_length = None
    if isinstance(abi_type, tuple):
        abi_type, dynamic_length = abi_type

    abi_type_str = str(abi.type_spec_from_annotation(abi_type))
    roundtrip_or_none = None
    if abi_type_str not in BAD_TYPES:
        roundtrip_or_none = ABIRoundtrip(
            abi.make(abi_type), length=dynamic_length
        ).pytealer()

    return abi_type, abi_type_str, dynamic_length, roundtrip_or_none


def test_abi_types_comprehensive():
    top_level_names = set()
    for at in ABI_TYPES:
        at = at[0] if isinstance(at, tuple) else at
        tli: str

        tli = (
            str(abi.NamedTuple)
            if isclass(at) and issubclass(at, abi.NamedTuple)
            else str(at)
        )
        tli = tli.split("[")[0] if tli.startswith("pyteal") else tli.split("'")[1]

        top_level_names.add(tli)

    def get_subclasses(cls):
        for subclass in cls.__subclasses__():
            yield from get_subclasses(subclass)
            yield subclass

    all_abi_names = {
        str(at).split("'")[1]
        for at in (
            cls
            for cls in abi.__dict__.values()
            if isinstance(cls, type)
            and issubclass(cls, abi.BaseType)
            and not cls.__abstractmethods__
            and cls is not abi.Tuple
        )
    }

    missing_cases = all_abi_names - top_level_names
    assert not missing_cases, f"missing round trip tests for {missing_cases}"


@pytest.mark.parametrize("abi_type", ABI_TYPES)
def test_pure_compilation(abi_type):
    print(f"Pure Compilation Test for {abi_type=}")
    abi_type, type_str, dynamic_length, roundtripper = roundtrip_setup(abi_type)

    if type_str in BAD_TYPES:
        print(
            f"Skipping encoding roundtrip test of '{abi_type}' because of {BAD_TYPES[type_str]}"
        )
        return

    sdk_abi_type = abi.algosdk_from_annotation(abi_type)

    abi_arg_types = roundtripper.abi_argument_types()
    abi_ret_type = roundtripper.abi_return_type()
    assert [sdk_abi_type] == abi_arg_types
    assert algosdk.abi.TupleType([sdk_abi_type] * 3) == abi_ret_type

    teal = roundtripper.compile(version=6)

    filename = (
        f"app_roundtrip_{sdk_abi_type}"
        + ("" if dynamic_length is None else f"_{dynamic_length}")
        + ".teal"
    )
    tealdir = GENERATED / "roundtrip"
    tealdir.mkdir(parents=True, exist_ok=True)

    save_to = tealdir / filename
    with open(save_to, "w") as f:
        f.write(teal)

    assert_teal_as_expected(save_to, FIXTURES / "roundtrip" / filename)


@pytest.mark.parametrize("abi_type", ABI_TYPES)
def test_roundtrip(abi_type):
    print(f"Round Trip Test for {abi_type=}")

    _, type_str, dynamic_length, roundtripper = roundtrip_setup(abi_type)

    if type_str in BAD_TYPES:
        print(
            f"Skipping encoding roundtrip test of '{abi_type}' because of {BAD_TYPES[type_str]}"
        )
        return

    sdk_abi_types = roundtripper.abi_argument_types()
    sdk_ret_type = roundtripper.abi_return_type()

    sdk_abi_str = str(sdk_abi_types[0])
    if sdk_abi_str in BAD_TEALS:
        print(
            f"Skipping encoding roundtrip test of '{sdk_abi_str}' because of {BAD_TEALS[sdk_abi_str]}"
        )
        return

    abi_strat = ABIStrategy(sdk_abi_types[0], dynamic_length=dynamic_length)
    rand_abi_instance = abi_strat.get_random()
    args = (rand_abi_instance,)
    inspector = roundtripper.dryrun(args)

    cost = inspector.cost()
    passed = inspector.passed()
    original, mut, mut_mut = inspector.last_log()

    print(
        f"""
{abi_type=}
{sdk_abi_str=}
{dynamic_length=}
{sdk_abi_types=}
{sdk_ret_type=}
{rand_abi_instance=}
{cost=}
{original=}
{mut=}
{mut_mut=}
"""
    )

    last_steps = 2

    assert passed == (cost <= 700), inspector.report(
        args, f"passed={passed} contradicted cost={cost}", last_steps=last_steps
    )
    assert rand_abi_instance == original, inspector.report(
        args, "rand_abi_instance v. original", last_steps=last_steps
    )
    assert original == mut_mut, inspector.report(
        args, "orginal v. mut_mut", last_steps=last_steps
    )

    expected_mut = abi_strat.mutate_for_roundtrip(rand_abi_instance)
    assert expected_mut == mut, inspector.report(
        args, "expected_mut v. mut", last_steps=last_steps
    )



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *


def approval_program():
    on_creation = Seq(
        [
            App.globalPut(Bytes("Creator"), Txn.sender()),
            Assert(Txn.application_args.length() == Int(4)),
            App.globalPut(Bytes("RegBegin"), Btoi(Txn.application_args[0])),
            App.globalPut(Bytes("RegEnd"), Btoi(Txn.application_args[1])),
            App.globalPut(Bytes("VoteBegin"), Btoi(Txn.application_args[2])),
            App.globalPut(Bytes("VoteEnd"), Btoi(Txn.application_args[3])),
            Return(Int(1)),
        ]
    )

    is_creator = Txn.sender() == App.globalGet(Bytes("Creator"))

    get_vote_of_sender = App.localGetEx(Int(0), App.id(), Bytes("voted"))

    on_closeout = Seq(
        [
            get_vote_of_sender,
            If(
                And(
                    Global.round() <= App.globalGet(Bytes("VoteEnd")),
                    get_vote_of_sender.hasValue(),
                ),
                App.globalPut(
                    get_vote_of_sender.value(),
                    App.globalGet(get_vote_of_sender.value()) - Int(1),
                ),
            ),
            Return(Int(1)),
        ]
    )

    on_register = Return(
        And(
            Global.round() >= App.globalGet(Bytes("RegBegin")),
            Global.round() <= App.globalGet(Bytes("RegEnd")),
        )
    )

    choice = Txn.application_args[1]
    choice_tally = App.globalGet(choice)
    on_vote = Seq(
        [
            Assert(
                And(
                    Global.round() >= App.globalGet(Bytes("VoteBegin")),
                    Global.round() <= App.globalGet(Bytes("VoteEnd")),
                )
            ),
            get_vote_of_sender,
            If(get_vote_of_sender.hasValue(), Return(Int(0))),
            App.globalPut(choice, choice_tally + Int(1)),
            App.localPut(Int(0), Bytes("voted"), choice),
            Return(Int(1)),
        ]
    )

    program = Cond(
        [Txn.application_id() == Int(0), on_creation],
        [Txn.on_completion() == OnComplete.DeleteApplication, Return(is_creator)],
        [Txn.on_completion() == OnComplete.UpdateApplication, Return(is_creator)],
        [Txn.on_completion() == OnComplete.CloseOut, on_closeout],
        [Txn.on_completion() == OnComplete.OptIn, on_register],
        [Txn.application_args[0] == Bytes("vote"), on_vote],
    )

    return program


def clear_state_program():
    get_vote_of_sender = App.localGetEx(Int(0), App.id(), Bytes("voted"))
    program = Seq(
        [
            get_vote_of_sender,
            If(
                And(
                    Global.round() <= App.globalGet(Bytes("VoteEnd")),
                    get_vote_of_sender.hasValue(),
                ),
                App.globalPut(
                    get_vote_of_sender.value(),
                    App.globalGet(get_vote_of_sender.value()) - Int(1),
                ),
            ),
            Return(Int(1)),
        ]
    )

    return program


if __name__ == "__main__":
    with open("vote_approval.teal", "w") as f:
        compiled = compileTeal(approval_program(), mode=Mode.Application, version=2)
        f.write(compiled)

    with open("vote_clear_state.teal", "w") as f:
        compiled = compileTeal(clear_state_program(), mode=Mode.Application, version=2)
        f.write(compiled)



import pytest

import pyteal as pt

avm5Options = pt.CompileOptions(version=5)
avm6Options = pt.CompileOptions(version=6)


def test_gitxn_invalid():
    for ctor, e in [
        (
            lambda: pt.Gitxn[pt.MAX_GROUP_SIZE],
            pt.TealInputError,
        ),
        (
            lambda: pt.Gitxn[-1],
            pt.TealInputError,
        ),
    ]:
        with pytest.raises(e):
            ctor()


def test_gitxn_valid():
    for i in range(pt.MAX_GROUP_SIZE):
        pt.Gitxn[i].sender()


def test_gitxn_expr_invalid():
    for f, e in [
        (
            lambda: pt.GitxnExpr(pt.Int(1), pt.TxnField.sender),
            pt.TealInputError,
        ),
        (
            lambda: pt.GitxnExpr(1, pt.TxnField.sender).__teal__(avm5Options),
            pt.TealInputError,
        ),
    ]:
        with pytest.raises(e):
            f()


def test_gitxn_expr_valid():
    pt.GitxnExpr(1, pt.TxnField.sender).__teal__(avm6Options)


def test_gitxna_expr_invalid():
    for f, e in [
        (
            lambda: pt.GitxnaExpr("Invalid_type", pt.TxnField.application_args, 1),
            pt.TealInputError,
        ),
        (
            lambda: pt.GitxnaExpr(1, pt.TxnField.application_args, "Invalid_type"),
            pt.TealInputError,
        ),
        (
            lambda: pt.GitxnaExpr(
                0, pt.TxnField.application_args, pt.Assert(pt.Int(1))
            ),
            pt.TealTypeError,
        ),
        (
            lambda: pt.GitxnaExpr(0, pt.TxnField.application_args, 0).__teal__(
                avm5Options
            ),
            pt.TealInputError,
        ),
    ]:
        with pytest.raises(e):
            f()


def test_gitxna_valid():
    [
        e.__teal__(avm6Options)
        for e in [
            pt.GitxnaExpr(0, pt.TxnField.application_args, 1),
            pt.GitxnaExpr(0, pt.TxnField.application_args, pt.Int(1)),
        ]
    ]


# txn_test.py performs additional testing



from pyteal.compiler.compiler import (
    MAX_TEAL_VERSION,
    MIN_TEAL_VERSION,
    DEFAULT_TEAL_VERSION,
    MAX_PROGRAM_VERSION,
    MIN_PROGRAM_VERSION,
    DEFAULT_PROGRAM_VERSION,
    CompileOptions,
    compileTeal,
)

from pyteal.compiler.optimizer import OptimizeOptions

__all__ = [
    "MAX_TEAL_VERSION",
    "MIN_TEAL_VERSION",
    "DEFAULT_TEAL_VERSION",
    "MAX_PROGRAM_VERSION",
    "MIN_PROGRAM_VERSION",
    "DEFAULT_PROGRAM_VERSION",
    "CompileOptions",
    "compileTeal",
    "OptimizeOptions",
]



# abstract types
from pyteal.ast.expr import Expr

# basic types
from pyteal.ast.leafexpr import LeafExpr
from pyteal.ast.addr import Addr
from pyteal.ast.bytes import Bytes
from pyteal.ast.int import Int, EnumInt
from pyteal.ast.methodsig import MethodSignature

# properties
from pyteal.ast.arg import Arg
from pyteal.ast.txn import (
    TxnType,
    TxnField,
    TxnExpr,
    TxnaExpr,
    TxnArray,
    TxnObject,
    Txn,
)
from pyteal.ast.gtxn import GtxnExpr, GtxnaExpr, TxnGroup, Gtxn
from pyteal.ast.block import Block
from pyteal.ast.gaid import GeneratedID
from pyteal.ast.gitxn import Gitxn, GitxnExpr, GitxnaExpr, InnerTxnGroup
from pyteal.ast.gload import ImportScratchValue
from pyteal.ast.global_ import Global, GlobalField
from pyteal.ast.app import App, AppField, OnComplete, AppParam, AppParamObject
from pyteal.ast.asset import (
    AssetHolding,
    AssetHoldingObject,
    AssetParam,
    AssetParamObject,
)
from pyteal.ast.acct import AccountParam, AccountParamObject

# inner txns
from pyteal.ast.itxn import InnerTxnBuilder, InnerTxn, InnerTxnAction

# meta
from pyteal.ast.array import Array
from pyteal.ast.tmpl import Tmpl
from pyteal.ast.nonce import Nonce
from pyteal.ast.pragma import Pragma
from pyteal.ast.comment import Comment

# unary ops
from pyteal.ast.unaryexpr import (
    UnaryExpr,
    Btoi,
    Itob,
    Len,
    BitLen,
    Sha256,
    Sha512_256,
    Sha3_256,
    Keccak256,
    Not,
    BitwiseNot,
    Sqrt,
    Pop,
    Balance,
    MinBalance,
    BytesNot,
    BytesSqrt,
    BytesZero,
    Log,
)

# binary ops
from pyteal.ast.binaryexpr import (
    BinaryExpr,
    Minus,
    Div,
    Mod,
    Exp,
    BitwiseAnd,
    BitwiseOr,
    BitwiseXor,
    ShiftLeft,
    ShiftRight,
    Eq,
    Neq,
    Lt,
    Le,
    Gt,
    Ge,
    GetBit,
    GetByte,
    BytesAdd,
    BytesMinus,
    BytesDiv,
    BytesMul,
    BytesMod,
    BytesAnd,
    BytesOr,
    BytesXor,
    BytesEq,
    BytesNeq,
    BytesLt,
    BytesLe,
    BytesGt,
    BytesGe,
    ExtractUint16,
    ExtractUint32,
    ExtractUint64,
)
from pyteal.ast.base64decode import Base64Decode

# ternary ops
from pyteal.ast.ternaryexpr import (
    Divw,
    Ed25519Verify,
    Ed25519Verify_Bare,
    SetBit,
    SetByte,
)
from pyteal.ast.substring import Substring, Extract, Suffix
from pyteal.ast.replace import Replace
from pyteal.ast.jsonref import JsonRef

# quaternary ops
from pyteal.ast.vrfverify import VrfVerify

# more ops
from pyteal.ast.naryexpr import NaryExpr, Add, And, Mul, Or, Concat
from pyteal.ast.widemath import WideRatio

# control flow
from pyteal.ast.if_ import If
from pyteal.ast.cond import Cond
from pyteal.ast.seq import Seq
from pyteal.ast.assert_ import Assert
from pyteal.ast.err import Err
from pyteal.ast.return_ import Return, Approve, Reject
from pyteal.ast.subroutine import (
    Subroutine,
    SubroutineDefinition,
    SubroutineDeclaration,
    SubroutineCall,
    SubroutineFnWrapper,
    ABIReturnSubroutine,
)
from pyteal.ast.while_ import While
from pyteal.ast.for_ import For
from pyteal.ast.break_ import Break
from pyteal.ast.continue_ import Continue

# misc
from pyteal.ast.scratch import (
    ScratchIndex,
    ScratchLoad,
    ScratchSlot,
    ScratchStackStore,
    ScratchStore,
)
from pyteal.ast.scratchvar import DynamicScratchVar, ScratchVar
from pyteal.ast.maybe import MaybeValue
from pyteal.ast.multi import MultiValue
from pyteal.ast.opup import OpUp, OpUpMode, OpUpFeeSource
from pyteal.ast.ecdsa import EcdsaCurve, EcdsaVerify, EcdsaDecompress, EcdsaRecover
from pyteal.ast.router import (
    Router,
    CallConfig,
    MethodConfig,
    OnCompleteAction,
    BareCallActions,
)

# abi
import pyteal.ast.abi as abi  # noqa: I250

__all__ = [
    "Expr",
    "LeafExpr",
    "Addr",
    "Bytes",
    "Int",
    "EnumInt",
    "MethodSignature",
    "Arg",
    "TxnType",
    "TxnField",
    "TxnExpr",
    "TxnaExpr",
    "TxnArray",
    "TxnObject",
    "Txn",
    "GtxnExpr",
    "GtxnaExpr",
    "TxnGroup",
    "Gtxn",
    "Block",
    "GeneratedID",
    "ImportScratchValue",
    "Global",
    "GlobalField",
    "App",
    "AppField",
    "OnComplete",
    "AppParam",
    "AppParamObject",
    "AssetHolding",
    "AssetHoldingObject",
    "AssetParam",
    "AssetParamObject",
    "AccountParam",
    "AccountParamObject",
    "InnerTxnBuilder",
    "InnerTxn",
    "InnerTxnAction",
    "Gitxn",
    "GitxnExpr",
    "GitxnaExpr",
    "InnerTxnGroup",
    "Array",
    "Tmpl",
    "Nonce",
    "Pragma",
    "Comment",
    "UnaryExpr",
    "Btoi",
    "Itob",
    "Len",
    "BitLen",
    "Sha256",
    "Sha512_256",
    "Sha3_256",
    "Keccak256",
    "Not",
    "BitwiseNot",
    "Sqrt",
    "Pop",
    "Balance",
    "MinBalance",
    "BinaryExpr",
    "Add",
    "Minus",
    "Mul",
    "Div",
    "Mod",
    "Exp",
    "Divw",
    "BitwiseAnd",
    "BitwiseOr",
    "BitwiseXor",
    "ShiftLeft",
    "ShiftRight",
    "Eq",
    "Neq",
    "Lt",
    "Le",
    "Gt",
    "Ge",
    "GetBit",
    "GetByte",
    "Ed25519Verify",
    "Ed25519Verify_Bare",
    "Substring",
    "Extract",
    "Suffix",
    "SetBit",
    "SetByte",
    "NaryExpr",
    "And",
    "Or",
    "Concat",
    "WideRatio",
    "If",
    "Cond",
    "Seq",
    "Assert",
    "Err",
    "Return",
    "Approve",
    "Reject",
    "Subroutine",
    "SubroutineDefinition",
    "SubroutineDeclaration",
    "SubroutineCall",
    "SubroutineFnWrapper",
    "ABIReturnSubroutine",
    "ScratchIndex",
    "ScratchLoad",
    "ScratchSlot",
    "ScratchStackStore",
    "ScratchStore",
    "DynamicScratchVar",
    "ScratchVar",
    "MaybeValue",
    "MultiValue",
    "OpUp",
    "OpUpMode",
    "OpUpFeeSource",
    "BytesAdd",
    "BytesMinus",
    "BytesDiv",
    "BytesMul",
    "BytesMod",
    "BytesAnd",
    "BytesOr",
    "BytesXor",
    "BytesEq",
    "BytesNeq",
    "BytesLt",
    "BytesLe",
    "BytesGt",
    "BytesGe",
    "BytesNot",
    "BytesSqrt",
    "BytesZero",
    "ExtractUint16",
    "ExtractUint32",
    "ExtractUint64",
    "Replace",
    "Base64Decode",
    "Log",
    "While",
    "For",
    "Break",
    "Continue",
    "Router",
    "CallConfig",
    "MethodConfig",
    "OnCompleteAction",
    "BareCallActions",
    "abi",
    "EcdsaCurve",
    "EcdsaVerify",
    "EcdsaDecompress",
    "EcdsaRecover",
    "JsonRef",
    "VrfVerify",
]



import pytest

import pyteal as pt
from pyteal.ast.maybe_test import assert_MaybeValue_equality

avm2Options = pt.CompileOptions()
avm4Options = pt.CompileOptions(version=4)
avm5Options = pt.CompileOptions(version=5)


def test_asset_holding_balance():
    args = pt.Int(0), pt.Int(17)
    expr = pt.AssetHolding.balance(args[0], args[1])
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 0),
            pt.TealOp(args[1], pt.Op.int, 17),
            pt.TealOp(expr, pt.Op.asset_holding_get, "AssetBalance"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_holding_balance_direct_ref():
    args = [pt.Txn.sender(), pt.Txn.assets[17]]
    expr = pt.AssetHolding.balance(args[0], args[1])
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.txn, "Sender"),
            pt.TealOp(args[1], pt.Op.txna, "Assets", 17),
            pt.TealOp(expr, pt.Op.asset_holding_get, "AssetBalance"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_holding_balance_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetHolding.balance(pt.Txn.sender(), pt.Bytes("100"))

    with pytest.raises(pt.TealTypeError):
        pt.AssetHolding.balance(pt.Int(0), pt.Txn.receiver())


def test_asset_holding_frozen():
    args = [pt.Int(0), pt.Int(17)]
    expr = pt.AssetHolding.frozen(args[0], args[1])
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 0),
            pt.TealOp(args[1], pt.Op.int, 17),
            pt.TealOp(expr, pt.Op.asset_holding_get, "AssetFrozen"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_holding_frozen_direct_ref():
    args = [pt.Txn.sender(), pt.Txn.assets[17]]
    expr = pt.AssetHolding.frozen(args[0], args[1])
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.txn, "Sender"),
            pt.TealOp(args[1], pt.Op.txna, "Assets", 17),
            pt.TealOp(expr, pt.Op.asset_holding_get, "AssetFrozen"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_holding_frozen_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetHolding.frozen(pt.Txn.sender(), pt.Bytes("17"))

    with pytest.raises(pt.TealTypeError):
        pt.AssetHolding.frozen(pt.Int(0), pt.Txn.receiver())


def test_asset_param_total():
    arg = pt.Int(0)
    expr = pt.AssetParam.total(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetTotal"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_total_direct_ref():
    arg = pt.Txn.assets[0]
    expr = pt.AssetParam.total(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetTotal"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_total_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.total(pt.Txn.sender())


def test_asset_param_decimals():
    arg = pt.Int(0)
    expr = pt.AssetParam.decimals(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetDecimals"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_decimals_direct_ref():
    arg = pt.Txn.assets[0]
    expr = pt.AssetParam.decimals(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetDecimals"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_decimals_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.decimals(pt.Txn.sender())


def test_asset_param_default_frozen():
    arg = pt.Int(0)
    expr = pt.AssetParam.defaultFrozen(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetDefaultFrozen"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_default_frozen_direct_ref():
    arg = pt.Txn.assets[0]
    expr = pt.AssetParam.defaultFrozen(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetDefaultFrozen"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_default_frozen_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.defaultFrozen(pt.Txn.sender())


def test_asset_param_unit_name():
    arg = pt.Int(0)
    expr = pt.AssetParam.unitName(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetUnitName"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_unit_name_direct_ref():
    arg = pt.Txn.assets[0]
    expr = pt.AssetParam.unitName(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetUnitName"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_unit_name_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.unitName(pt.Txn.sender())


def test_asset_param_name():
    arg = pt.Int(0)
    expr = pt.AssetParam.name(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetName"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_name_direct_ref():
    arg = pt.Txn.assets[0]
    expr = pt.AssetParam.name(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetName"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_name_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.name(pt.Txn.sender())


def test_asset_param_url():
    arg = pt.Int(0)
    expr = pt.AssetParam.url(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetURL"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_url_direct_ref():
    arg = pt.Txn.assets[0]
    expr = pt.AssetParam.url(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetURL"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_url_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.url(pt.Txn.sender())


def test_asset_param_metadata_hash():
    arg = pt.Int(0)
    expr = pt.AssetParam.metadataHash(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetMetadataHash"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_metadata_hash_direct_ref():
    arg = pt.Txn.assets[0]
    expr = pt.AssetParam.metadataHash(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetMetadataHash"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_metadata_hash_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.metadataHash(pt.Txn.sender())


def test_asset_param_manager():
    arg = pt.Int(0)
    expr = pt.AssetParam.manager(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetManager"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_manager_direct_ref():
    arg = pt.Txn.assets[0]
    expr = pt.AssetParam.manager(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetManager"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_manager_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.manager(pt.Txn.sender())


def test_asset_param_reserve():
    arg = pt.Int(2)
    expr = pt.AssetParam.reserve(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 2),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetReserve"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_reserve_direct_ref():
    arg = pt.Txn.assets[2]
    expr = pt.AssetParam.reserve(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 2),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetReserve"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_reserve_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.reserve(pt.Txn.sender())


def test_asset_param_freeze():
    arg = pt.Int(0)
    expr = pt.AssetParam.freeze(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetFreeze"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_freeze_direct_ref():
    arg = pt.Txn.assets[0]
    expr = pt.AssetParam.freeze(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 0),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetFreeze"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_freeze_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.freeze(pt.Txn.sender())


def test_asset_param_clawback():
    arg = pt.Int(1)
    expr = pt.AssetParam.clawback(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetClawback"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_clawback_direct_ref():
    arg = pt.Txn.assets[1]
    expr = pt.AssetParam.clawback(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.txna, "Assets", 1),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetClawback"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_clawback_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.clawback(pt.Txn.sender())


def test_asset_param_creator_valid():
    arg = pt.Int(1)
    expr = pt.AssetParam.creator(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.asset_params_get, "AssetCreator"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_asset_param_creator_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AssetParam.creator(pt.Txn.sender())


def test_AssetHoldingObject():
    for asset in (pt.Int(1), pt.Int(100)):
        for account in (
            pt.Int(7),
            pt.Addr("QSA6K5MNJPEGO5SDSWXBM3K4UEI3Q2NCPS2OUXVJI5QPCHMVI27MFRSHKI"),
        ):
            obj = pt.AssetHoldingObject(asset, account)

            assert obj._asset is asset
            assert obj._account is account

            assert_MaybeValue_equality(
                obj.balance(), pt.AssetHolding.balance(account, asset), avm5Options
            )
            assert_MaybeValue_equality(
                obj.frozen(), pt.AssetHolding.frozen(account, asset), avm5Options
            )


def test_AssetParamObject():
    for asset in (pt.Int(1), pt.Int(100)):
        obj = pt.AssetParamObject(asset)

        assert obj._asset is asset

        assert_MaybeValue_equality(obj.total(), pt.AssetParam.total(asset), avm5Options)
        assert_MaybeValue_equality(
            obj.decimals(), pt.AssetParam.decimals(asset), avm5Options
        )
        assert_MaybeValue_equality(
            obj.default_frozen(), pt.AssetParam.defaultFrozen(asset), avm5Options
        )
        assert_MaybeValue_equality(
            obj.unit_name(), pt.AssetParam.unitName(asset), avm5Options
        )
        assert_MaybeValue_equality(obj.name(), pt.AssetParam.name(asset), avm5Options)
        assert_MaybeValue_equality(obj.url(), pt.AssetParam.url(asset), avm5Options)
        assert_MaybeValue_equality(
            obj.metadata_hash(), pt.AssetParam.metadataHash(asset), avm5Options
        )
        assert_MaybeValue_equality(
            obj.manager_address(), pt.AssetParam.manager(asset), avm5Options
        )
        assert_MaybeValue_equality(
            obj.reserve_address(), pt.AssetParam.reserve(asset), avm5Options
        )
        assert_MaybeValue_equality(
            obj.freeze_address(), pt.AssetParam.freeze(asset), avm5Options
        )
        assert_MaybeValue_equality(
            obj.clawback_address(), pt.AssetParam.clawback(asset), avm5Options
        )
        assert_MaybeValue_equality(
            obj.creator_address(), pt.AssetParam.creator(asset), avm5Options
        )



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *


""" Template for layer 1 dutch auction (from Fabrice and Shai)
"""

tmpl_start_round = Tmpl.Int("TMPL_START_ROUND")
tmpl_start_price = Tmpl.Int("TMPL_START_PRICE")
tmpl_supply = Tmpl.Int("TMPL_N")
tmpl_price_decrement = Tmpl.Int("TMPL_PRICE_INCREMENT")
tmpl_period = Tmpl.Int("TMPL_PERIOD")
tmpl_asset_a = Tmpl.Int("TMPL_ASSET_A")
tmpl_asset_b = Tmpl.Int("TMPL_ASSET_B")
tmpl_asset_c = Tmpl.Int("TMPL_ASSET_C")
tmpl_asset_d = Tmpl.Int("TMPL_ASSET_D")
tmpl_receiver = Tmpl.Addr("TMPL_RECEIVER")
tmpl_algo_sink = Tmpl.Addr("TMPL_ALGO_ZERO")
tmpl_asset_a_sink = Tmpl.Addr("TMPL_A_ZERO")
tmpl_asset_b_sink = Tmpl.Addr("TMPL_B_ZERO")
tmpl_asset_c_sink = Tmpl.Addr("TMPL_C_ZERO")
tmpl_asset_d_sink = Tmpl.Addr("TMPL_D_ZERO")
tmpl_redeem_round = Tmpl.Int("TMPL_REDEEM_ROUND")
tmpl_wrapup_time = Tmpl.Int("TMPL_WRAPUP_ROUND")


def dutch_auction(
    start_round=tmpl_start_round,
    start_price=tmpl_start_price,
    supply=tmpl_supply,
    price_decrement=tmpl_price_decrement,
    period=tmpl_period,
    asset_a=tmpl_asset_a,
    asset_b=tmpl_asset_b,
    asset_c=tmpl_asset_c,
    asset_d=tmpl_asset_d,
    receiver=tmpl_receiver,
    algo_sink=tmpl_algo_sink,
    asset_a_sink=tmpl_asset_a_sink,
    asset_b_sink=tmpl_asset_b_sink,
    asset_c_sink=tmpl_asset_c_sink,
    asset_d_sink=tmpl_asset_d_sink,
    redeem_round=tmpl_redeem_round,
    wrapup_time=tmpl_wrapup_time,
):
    # the definition of i simplifies the original desgin by only constraining last_valid here
    i_upper = (Gtxn[0].last_valid() - start_round) / period

    bid = And(
        Gtxn[0].rekey_to() == Global.zero_address(),
        Gtxn[1].rekey_to() == Global.zero_address(),
        Gtxn[2].rekey_to() == Global.zero_address(),
        Gtxn[3].rekey_to() == Global.zero_address(),
        Gtxn[4].rekey_to() == Global.zero_address(),
        Gtxn[0].last_valid() == Gtxn[1].last_valid(),
        Gtxn[1].last_valid() == Gtxn[2].last_valid(),
        Gtxn[2].last_valid() == Gtxn[3].last_valid(),
        Gtxn[3].last_valid() == Gtxn[4].last_valid(),
        Gtxn[4].last_valid() < (start_round + i_upper * period),
        Gtxn[0].type_enum() == TxnType.AssetTransfer,
        Gtxn[0].xfer_asset() == asset_d,
        Gtxn[0].receiver() == receiver,
        Gtxn[1].type_enum() == TxnType.AssetTransfer,
        Gtxn[1].xfer_asset() == asset_b,
        Gtxn[2].type_enum() == TxnType.AssetTransfer,
        Gtxn[2].xfer_asset() == asset_c,
        Gtxn[3].type_enum() == TxnType.AssetTransfer,
        Gtxn[4].xfer_asset() == asset_c,
        Gtxn[4].type_enum() == TxnType.Payment,
        Gtxn[4].amount()
        == (Gtxn[1].fee() + Gtxn[2].fee() + Gtxn[3].fee()),  # ? why only 1, 2, 3
        Gtxn[0].asset_amount() == Gtxn[1].asset_amount(),
        Gtxn[1].asset_amount() == Gtxn[2].asset_amount(),
        Gtxn[3].asset_amount() == i_upper * supply * price_decrement,
        Gtxn[0].sender() == Gtxn[1].receiver(),
        Gtxn[2].receiver() == asset_c_sink,
        Gtxn[1].sender() == Gtxn[2].sender(),
        Gtxn[2].sender() == Gtxn[3].sender(),
        Gtxn[3].sender() == Gtxn[3].receiver(),
        Gtxn[3].receiver() == Gtxn[4].receiver(),
    )

    redeem = And(
        Gtxn[0].rekey_to() == Global.zero_address(),
        Gtxn[1].rekey_to() == Global.zero_address(),
        Gtxn[2].rekey_to() == Global.zero_address(),
        Gtxn[3].rekey_to() == Global.zero_address(),
        Gtxn[0].first_valid() == Gtxn[1].first_valid(),
        Gtxn[1].first_valid() == Gtxn[2].first_valid(),
        Gtxn[2].first_valid() == Gtxn[3].first_valid(),
        Gtxn[3].first_valid() >= redeem_round,
        Gtxn[0].type_enum() == TxnType.AssetTransfer,
        Gtxn[0].xfer_asset() == asset_b,
        Gtxn[1].type_enum() == TxnType.AssetTransfer,
        Gtxn[1].xfer_asset() == asset_a,
        Gtxn[2].type_enum() == TxnType.AssetTransfer,
        Gtxn[2].xfer_asset() == asset_c,
        Gtxn[2].asset_amount() == redeem_round * supply * price_decrement,
        Gtxn[3].type_enum() == TxnType.Payment,
        Gtxn[3].amount() == Gtxn[1].fee() + Gtxn[2].fee(),
        Gtxn[1].asset_amount()
        == Gtxn[0].asset_amount() * (start_price - price_decrement * Btoi(Arg(0))),
        Gtxn[0].sender() == Gtxn[1].receiver(),
        Gtxn[0].receiver() == Gtxn[1].sender(),
        Gtxn[1].sender() == Gtxn[2].sender(),
        Gtxn[2].sender() == Gtxn[2].receiver(),
        Gtxn[2].receiver() == Gtxn[3].receiver(),
    )

    wrapup = And(
        Txn.rekey_to() == Global.zero_address(),
        Txn.first_valid() >= start_round + wrapup_time,
        Or(
            And(
                Txn.type_enum() == TxnType.Payment,
                Txn.amount() == Int(0),
                Txn.close_remainder_to() == receiver,
            ),
            And(
                Txn.type_enum() == TxnType.AssetTransfer,
                Txn.asset_amount() == Int(0),
                Txn.asset_close_to() == receiver,
                Txn.xfer_asset() == asset_a,
            ),
        ),
    )

    dutch = Cond(
        [Global.group_size() == Int(5), bid],
        [Global.group_size() == Int(4), redeem],
        [Global.group_size() == Int(1), wrapup],
    )

    return dutch


if __name__ == "__main__":
    print(compileTeal(dutch_auction(), mode=Mode.Signature, version=2))



import pytest

import pyteal as pt

avm6Options = pt.CompileOptions(version=6)
avm7Options = pt.CompileOptions(version=7)


def test_replace_immediate():
    args = [pt.Bytes("my string"), pt.Int(0), pt.Bytes("abcdefghi")]
    expr = pt.Replace(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"my string"'),
            pt.TealOp(args[2], pt.Op.byte, '"abcdefghi"'),
            pt.TealOp(expr, pt.Op.replace2, 0),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_replace_stack_int():
    my_string = "*" * 257
    args = [pt.Bytes(my_string), pt.Int(256), pt.Bytes("ab")]
    expr = pt.Replace(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(args[1], pt.Op.int, 256),
            pt.TealOp(args[2], pt.Op.byte, '"ab"'),
            pt.TealOp(expr, pt.Op.replace3),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


# Mirrors `test_replace_stack_int`, but attempts replacement with start != pt.Int.
def test_replace_stack_not_int():
    my_string = "*" * 257
    add = pt.Add(pt.Int(254), pt.Int(2))
    args = [pt.Bytes(my_string), add, pt.Bytes("ab")]
    expr = pt.Replace(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(pt.Int(254), pt.Op.int, 254),
            pt.TealOp(pt.Int(2), pt.Op.int, 2),
            pt.TealOp(add, pt.Op.add),
            pt.TealOp(args[2], pt.Op.byte, '"ab"'),
            pt.TealOp(expr, pt.Op.replace3),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_replace_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Replace(pt.Bytes("my string"), pt.Int(0), pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        pt.Replace(
            pt.Bytes("my string"), pt.Bytes("should be int"), pt.Bytes("abcdefghi")
        )

    with pytest.raises(pt.TealTypeError):
        pt.Replace(pt.Bytes("my string"), pt.Txn.sender(), pt.Bytes("abcdefghi"))



import pytest
import pyteal as pt
from pyteal import abi

from pyteal.ast.abi.address import AddressLength
from pyteal.ast.abi.type_test import ContainerType
from pyteal.ast.abi.util import substring_for_decoding


options = pt.CompileOptions(version=5)


def test_AddressTypeSpec_str():
    assert str(abi.AddressTypeSpec()) == "address"


def test_AddressTypeSpec_is_dynamic():
    assert (abi.AddressTypeSpec()).is_dynamic() is False


def test_AddressTypeSpec_byte_length_static():
    assert (abi.AddressTypeSpec()).byte_length_static() == abi.AddressLength.Bytes


def test_AddressTypeSpec_length_static():
    assert (abi.AddressTypeSpec()).length_static() == abi.AddressLength.Bytes


def test_AddressTypeSpec_new_instance():
    assert isinstance(abi.AddressTypeSpec().new_instance(), abi.Address)


def test_AddressTypeSpec_eq():
    assert abi.AddressTypeSpec() == abi.AddressTypeSpec()

    for otherType in (
        abi.ByteTypeSpec(),
        abi.StaticArrayTypeSpec(abi.ByteTypeSpec(), 32),
        abi.DynamicArrayTypeSpec(abi.ByteTypeSpec()),
    ):
        assert abi.AddressTypeSpec() != otherType


def test_Address_encode():
    value = abi.Address()
    expr = value.encode()
    assert expr.type_of() == pt.TealType.bytes
    assert expr.has_return() is False

    expected = pt.TealSimpleBlock(
        [pt.TealOp(expr, pt.Op.load, value.stored_value.slot)]
    )
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_Address_decode():
    address = bytes([0] * abi.AddressLength.Bytes)
    encoded = pt.Bytes(address)

    for start_index in (None, pt.Int(0)):
        for end_index in (None, pt.Int(1)):
            for length in (None, pt.Int(2)):
                value = abi.Address()

                if end_index is not None and length is not None:
                    with pytest.raises(pt.TealInputError):
                        value.decode(
                            encoded,
                            start_index=start_index,
                            end_index=end_index,
                            length=length,
                        )
                    continue

                expr = value.decode(
                    encoded, start_index=start_index, end_index=end_index, length=length
                )
                assert expr.type_of() == pt.TealType.none
                assert expr.has_return() is False

                expectedExpr = value.stored_value.store(
                    substring_for_decoding(
                        encoded,
                        start_index=start_index,
                        end_index=end_index,
                        length=length,
                    )
                )
                expected, _ = expectedExpr.__teal__(options)
                expected.addIncoming()
                expected = pt.TealBlock.NormalizeBlocks(expected)

                actual, _ = expr.__teal__(options)
                actual.addIncoming()
                actual = pt.TealBlock.NormalizeBlocks(actual)

                with pt.TealComponent.Context.ignoreExprEquality():
                    assert actual == expected


def test_Address_get():
    value = abi.Address()
    expr = value.get()
    assert expr.type_of() == pt.TealType.bytes
    assert expr.has_return() is False

    expected = pt.TealSimpleBlock(
        [pt.TealOp(expr, pt.Op.load, value.stored_value.slot)]
    )
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_Address_set_StaticArray():
    value_to_set = abi.StaticArray(
        abi.StaticArrayTypeSpec(abi.ByteTypeSpec(), abi.AddressLength.Bytes)
    )
    value = abi.Address()
    expr = value.set(value_to_set)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.load, value_to_set.stored_value.slot),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        bogus = abi.StaticArray(abi.StaticArrayTypeSpec(abi.ByteTypeSpec(), 10))
        value.set(bogus)


def test_Address_set_str():
    for value_to_set in ("CEZZTYHNTVIZFZWT6X2R474Z2P3Q2DAZAKIRTPBAHL3LZ7W4O6VBROVRQA",):
        value = abi.Address()
        expr = value.set(value_to_set)
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(None, pt.Op.addr, value_to_set),
                pt.TealOp(None, pt.Op.store, value.stored_value.slot),
            ]
        )

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected

        with pytest.raises(pt.TealInputError):
            value.set(" " * 16)


def test_Address_set_bytes():
    for value_to_set in (bytes(32),):
        value = abi.Address()
        expr = value.set(value_to_set)
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(None, pt.Op.byte, f"0x{value_to_set.hex()}"),
                pt.TealOp(None, pt.Op.store, value.stored_value.slot),
            ]
        )

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected

        with pytest.raises(pt.TealInputError):
            value.set(bytes(16))

        with pytest.raises(pt.TealInputError):
            value.set(16)


def test_Address_set_expr():
    for value_to_set in [pt.Global(pt.GlobalField.zero_address)]:
        value = abi.Address()
        expr = value.set(value_to_set)
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        vts, _ = value_to_set.__teal__(options)
        expected = pt.TealSimpleBlock(
            [
                vts.ops[0],
                pt.TealOp(None, pt.Op.store, value.stored_value.slot),
                pt.TealOp(None, pt.Op.load, value.stored_value.slot),
                pt.TealOp(None, pt.Op.len),
                pt.TealOp(None, pt.Op.int, AddressLength.Bytes.value),
                pt.TealOp(None, pt.Op.eq),
                pt.TealOp(None, pt.Op.assert_),
            ]
        )

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected


def test_Address_set_copy():
    value = abi.Address()
    other = abi.Address()
    expr = value.set(other)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.load, other.stored_value.slot),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        value.set(abi.String())


def test_Address_set_computed():
    av = pt.Addr("MDDKJUCTY57KA2PBFI44CLTJ5YHY5YVS4SVQUPZAWSRV2ZAVFKI33O6YPE")
    computed_value = ContainerType(abi.AddressTypeSpec(), av)

    value = abi.Address()
    expr = value.set(computed_value)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    _, byte_ops = av.__teal__(options)
    expected = pt.TealSimpleBlock(
        [
            byte_ops.ops[0],
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        value.set(ContainerType(abi.ByteTypeSpec(), pt.Int(0x01)))



from typing import List, cast, TYPE_CHECKING
from pyteal.ast.seq import _use_seq_if_multiple

from pyteal.types import TealType, require_type
from pyteal.ir import TealOp, Op, TealSimpleBlock, TealConditionalBlock
from pyteal.errors import TealInputError
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


def _reformat_multi_argv(argv: tuple[list[Expr], ...]) -> list[list[Expr]]:
    """Reformat a list of lists of expressions with potentially multiple value expressions into a list of lists of expressions with only one value expression
        by using Seq blocks where appropriate.

    Example:
        [ [a, b, c], [d, e, f] ] -> [ [a, Seq(b, c)], [d, Seq(e, f)] ]
    """
    reformatted = []
    for arg in argv:
        # Note: this is not a valid Cond arg, but will be caught later in Cond.__init__()
        if len(arg) <= 1:
            reformatted.append(arg)
        else:
            reformatted.append([arg[0], _use_seq_if_multiple(arg[1:])])

    return reformatted


class Cond(Expr):
    """A chainable branching expression that supports an arbitrary number of conditions."""

    def __init__(self, *argv: List[Expr]):
        """Create a new Cond expression.

        At least one argument must be provided, and each argument must be a list with two or more elements.
        The first element is a condition which evalutes to uint64, and the remaining elements are the body
        of the condition, which will execute if that condition is true. The last elements of the condition bodies
        must have the same return type. During execution, each condition is tested in order, and the first
        condition to evaluate to a true value will cause its associated body to execute and become
        the value for this Cond expression. If no condition evalutes to a true value, the Cond
        expression produces an error and the TEAL program terminates.

        Example:
            .. code-block:: python

                Cond([Global.group_size() == Int(5), bid],
                    [Global.group_size() == Int(4), redeem, log],
                    [Global.group_size() == Int(1), wrapup])
        """
        super().__init__()

        if len(argv) < 1:
            raise TealInputError("Cond requires at least one [condition, value]")

        value_type = None
        sequenced_argv = _reformat_multi_argv(argv)

        for arg in sequenced_argv:
            msg = "Cond should be in the form of Cond([cond1, value1], [cond2, value2], ...), error in {}"
            if not isinstance(arg, list):
                raise TealInputError(msg.format(arg))
            if len(arg) != 2:
                raise TealInputError(msg.format(arg))

            require_type(arg[0], TealType.uint64)  # cond_n should be int

            if value_type is None:  # the types of all branches should be the same
                value_type = arg[1].type_of()
            else:
                require_type(arg[1], value_type)

        self.value_type = value_type
        self.args = sequenced_argv

    def __teal__(self, options: "CompileOptions"):
        start = None
        end = TealSimpleBlock([])
        prevBranch = None
        for i, (cond, pred) in enumerate(self.args):
            condStart, condEnd = cond.__teal__(options)
            predStart, predEnd = pred.__teal__(options)

            branchBlock = TealConditionalBlock([])
            branchBlock.setTrueBlock(predStart)

            condEnd.setNextBlock(branchBlock)
            predEnd.setNextBlock(end)
            if i == 0:
                start = condStart
            else:
                cast(TealConditionalBlock, prevBranch).setFalseBlock(condStart)
            prevBranch = branchBlock

        errBlock = TealSimpleBlock([TealOp(self, Op.err)])
        cast(TealConditionalBlock, prevBranch).setFalseBlock(errBlock)

        return start, end

    def __str__(self):
        ret_str = "(Cond"
        for a in self.args:
            ret_str += " [" + a[0].__str__() + ", " + a[1].__str__() + "]"
        ret_str += ")"
        return ret_str

    def type_of(self):
        return self.value_type

    def has_return(self):
        # this expression has a return op only if all possible conditions result in a return op
        return all(pred.has_return() for (_, pred) in self.args)


Cond.__module__ = "pyteal"



import pytest

import pyteal as pt
from pyteal.ast.maybe_test import assert_MaybeValue_equality

options = pt.CompileOptions()
avm4Options = pt.CompileOptions(version=4)
avm5Options = pt.CompileOptions(version=5)


def test_on_complete():
    assert pt.OnComplete.NoOp.__teal__(options)[0] == pt.TealSimpleBlock(
        [pt.TealOp(pt.OnComplete.NoOp, pt.Op.int, "NoOp")]
    )

    assert pt.OnComplete.OptIn.__teal__(options)[0] == pt.TealSimpleBlock(
        [pt.TealOp(pt.OnComplete.OptIn, pt.Op.int, "OptIn")]
    )

    assert pt.OnComplete.CloseOut.__teal__(options)[0] == pt.TealSimpleBlock(
        [pt.TealOp(pt.OnComplete.CloseOut, pt.Op.int, "CloseOut")]
    )

    assert pt.OnComplete.ClearState.__teal__(options)[0] == pt.TealSimpleBlock(
        [pt.TealOp(pt.OnComplete.ClearState, pt.Op.int, "ClearState")]
    )

    assert pt.OnComplete.UpdateApplication.__teal__(options)[0] == pt.TealSimpleBlock(
        [pt.TealOp(pt.OnComplete.UpdateApplication, pt.Op.int, "UpdateApplication")]
    )

    assert pt.OnComplete.DeleteApplication.__teal__(options)[0] == pt.TealSimpleBlock(
        [pt.TealOp(pt.OnComplete.DeleteApplication, pt.Op.int, "DeleteApplication")]
    )


def test_app_id():
    expr = pt.App.id()
    assert expr.type_of() == pt.TealType.uint64
    with pt.TealComponent.Context.ignoreExprEquality():
        assert (
            expr.__teal__(options)[0]
            == pt.Global.current_application_id().__teal__(options)[0]
        )


def test_opted_in():
    args = [pt.Int(1), pt.Int(12)]
    expr = pt.App.optedIn(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 12),
            pt.TealOp(expr, pt.Op.app_opted_in),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_opted_in_direct_ref():
    args = [pt.Bytes("sender address"), pt.Int(100)]
    expr = pt.App.optedIn(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"sender address"'),
            pt.TealOp(args[1], pt.Op.int, 100),
            pt.TealOp(expr, pt.Op.app_opted_in),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_opted_in_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.App.optedIn(pt.Bytes("sender"), pt.Bytes("100"))

    with pytest.raises(pt.TealTypeError):
        pt.App.optedIn(pt.Int(123456), pt.Bytes("364"))


def test_local_get():
    args = [pt.Int(0), pt.Bytes("key")]
    expr = pt.App.localGet(args[0], args[1])
    assert expr.type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 0),
            pt.TealOp(args[1], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.app_local_get),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_local_get_direct_ref():
    args = [pt.Txn.sender(), pt.Bytes("key")]
    expr = pt.App.localGet(args[0], args[1])
    assert expr.type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.txn, "Sender"),
            pt.TealOp(args[1], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.app_local_get),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_local_get_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.App.localGet(pt.Txn.sender(), pt.Int(1337))

    with pytest.raises(pt.TealTypeError):
        pt.App.localGet(pt.Int(0), pt.Int(1))


def test_local_get_ex():
    args = [pt.Int(0), pt.Int(6), pt.Bytes("key")]
    expr = pt.App.localGetEx(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 0),
            pt.TealOp(args[1], pt.Op.int, 6),
            pt.TealOp(args[2], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.app_local_get_ex),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_local_get_ex_direct_ref():
    args = [pt.Txn.sender(), pt.Int(6), pt.Bytes("key")]
    expr = pt.App.localGetEx(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.txn, "Sender"),
            pt.TealOp(args[1], pt.Op.int, 6),
            pt.TealOp(args[2], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.app_local_get_ex),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_local_get_ex_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), pt.Int(0x123456))

    with pytest.raises(pt.TealTypeError):
        pt.App.localGetEx(pt.Int(0), pt.Bytes("app"), pt.Bytes("key"))


def test_global_get():
    arg = pt.Bytes("key")
    expr = pt.App.globalGet(arg)
    assert expr.type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock(
        [pt.TealOp(arg, pt.Op.byte, '"key"'), pt.TealOp(expr, pt.Op.app_global_get)]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_global_get_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.App.globalGet(pt.Int(7))


def test_global_get_ex():
    args = [pt.Int(6), pt.Bytes("key")]
    expr = pt.App.globalGetEx(args[0], args[1])
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 6),
            pt.TealOp(args[1], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.app_global_get_ex),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_global_get_ex_direct_ref():
    args = [pt.Txn.applications[0], pt.Bytes("key")]
    expr = pt.App.globalGetEx(args[0], args[1])
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.txna, "Applications", 0),
            pt.TealOp(args[1], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.app_global_get_ex),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_global_get_ex_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.App.globalGetEx(pt.Bytes("app"), pt.Int(12))

    with pytest.raises(pt.TealTypeError):
        pt.App.globalGetEx(pt.Int(0), pt.Int(1))


def test_local_put():
    args = [pt.Int(0), pt.Bytes("key"), pt.Int(5)]
    expr = pt.App.localPut(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 0),
            pt.TealOp(args[1], pt.Op.byte, '"key"'),
            pt.TealOp(args[2], pt.Op.int, 5),
            pt.TealOp(expr, pt.Op.app_local_put),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_local_put_direct_ref():
    args = [pt.Txn.sender(), pt.Bytes("key"), pt.Int(5)]
    expr = pt.App.localPut(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.txn, "Sender"),
            pt.TealOp(args[1], pt.Op.byte, '"key"'),
            pt.TealOp(args[2], pt.Op.int, 5),
            pt.TealOp(expr, pt.Op.app_local_put),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_local_put_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.App.localPut(pt.Txn.sender(), pt.Int(55), pt.Int(5))

    with pytest.raises(pt.TealTypeError):
        pt.App.localPut(pt.Int(1), pt.Int(0), pt.Int(5))

    with pytest.raises(pt.TealTypeError):
        pt.App.localPut(pt.Int(1), pt.Bytes("key"), pt.Pop(pt.Int(1)))


def test_global_put():
    args = [pt.Bytes("key"), pt.Int(5)]
    expr = pt.App.globalPut(args[0], args[1])
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"key"'),
            pt.TealOp(args[1], pt.Op.int, 5),
            pt.TealOp(expr, pt.Op.app_global_put),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_global_put_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.App.globalPut(pt.Int(0), pt.Int(5))

    with pytest.raises(pt.TealTypeError):
        pt.App.globalPut(pt.Bytes("key"), pt.Pop(pt.Int(1)))


def test_local_del():
    args = [pt.Int(0), pt.Bytes("key")]
    expr = pt.App.localDel(args[0], args[1])
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 0),
            pt.TealOp(args[1], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.app_local_del),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_local_del_direct_ref():
    args = [pt.Txn.sender(), pt.Bytes("key")]
    expr = pt.App.localDel(args[0], args[1])
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.txn, "Sender"),
            pt.TealOp(args[1], pt.Op.byte, '"key"'),
            pt.TealOp(expr, pt.Op.app_local_del),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_local_del_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.App.localDel(pt.Txn.sender(), pt.Int(123))

    with pytest.raises(pt.TealTypeError):
        pt.App.localDel(pt.Int(1), pt.Int(2))


def test_global_del():
    arg = pt.Bytes("key")
    expr = pt.App.globalDel(arg)
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [pt.TealOp(arg, pt.Op.byte, '"key"'), pt.TealOp(expr, pt.Op.app_global_del)]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_global_del_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.App.globalDel(pt.Int(2))


def test_app_param_approval_program_valid():
    arg = pt.Int(1)
    expr = pt.AppParam.approvalProgram(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.app_params_get, "AppApprovalProgram"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_app_param_approval_program_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AppParam.approvalProgram(pt.Txn.sender())


def test_app_param_clear_state_program_valid():
    arg = pt.Int(0)
    expr = pt.AppParam.clearStateProgram(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.app_params_get, "AppClearStateProgram"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_app_param_clear_state_program_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AppParam.clearStateProgram(pt.Txn.sender())


def test_app_param_global_num_uint_valid():
    arg = pt.Int(1)
    expr = pt.AppParam.globalNumUint(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.app_params_get, "AppGlobalNumUint"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_app_param_global_num_uint_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AppParam.globalNumUint(pt.Txn.sender())


def test_app_param_global_num_byte_slice_valid():
    arg = pt.Int(1)
    expr = pt.AppParam.globalNumByteSlice(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.app_params_get, "AppGlobalNumByteSlice"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_app_param_global_num_byte_slice_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AppParam.globalNumByteSlice(pt.Txn.sender())


def test_app_param_local_num_uint_valid():
    arg = pt.Int(1)
    expr = pt.AppParam.localNumUint(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.app_params_get, "AppLocalNumUint"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_app_param_local_num_uint_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AppParam.localNumUint(pt.Txn.sender())


def test_app_param_local_num_byte_slice_valid():
    arg = pt.Int(1)
    expr = pt.AppParam.localNumByteSlice(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.app_params_get, "AppLocalNumByteSlice"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_app_param_local_num_byte_slice_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AppParam.localNumByteSlice(pt.Txn.sender())


def test_app_param_extra_programs_page_valid():
    arg = pt.Int(1)
    expr = pt.AppParam.extraProgramPages(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.app_params_get, "AppExtraProgramPages"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_app_param_extra_program_pages_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AppParam.extraProgramPages(pt.Txn.sender())


def test_app_param_creator_valid():
    arg = pt.Int(1)
    expr = pt.AppParam.creator(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.app_params_get, "AppCreator"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_app_param_creator_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AppParam.creator(pt.Txn.sender())


def test_app_param_address_valid():
    arg = pt.Int(1)
    expr = pt.AppParam.address(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.app_params_get, "AppAddress"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_app_param_address_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.AppParam.address(pt.Txn.sender())


def test_AppParamObject():
    for app in (pt.Int(1), pt.Int(100)):
        obj = pt.AppParamObject(app)

        assert obj._app is app

        assert_MaybeValue_equality(
            obj.approval_program(), pt.AppParam.approvalProgram(app), avm5Options
        )
        assert_MaybeValue_equality(
            obj.clear_state_program(), pt.AppParam.clearStateProgram(app), avm5Options
        )
        assert_MaybeValue_equality(
            obj.global_num_uint(), pt.AppParam.globalNumUint(app), avm5Options
        )
        assert_MaybeValue_equality(
            obj.global_num_byte_slice(),
            pt.AppParam.globalNumByteSlice(app),
            avm5Options,
        )
        assert_MaybeValue_equality(
            obj.local_num_uint(), pt.AppParam.localNumUint(app), avm5Options
        )
        assert_MaybeValue_equality(
            obj.local_num_byte_slice(), pt.AppParam.localNumByteSlice(app), avm5Options
        )
        assert_MaybeValue_equality(
            obj.extra_program_pages(), pt.AppParam.extraProgramPages(app), avm5Options
        )
        assert_MaybeValue_equality(
            obj.creator_address(), pt.AppParam.creator(app), avm5Options
        )
        assert_MaybeValue_equality(obj.address(), pt.AppParam.address(app), avm5Options)



from typing import Callable, List, Union, TYPE_CHECKING

from pyteal.types import TealType
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr
from pyteal.ast.leafexpr import LeafExpr
from pyteal.ast.scratch import ScratchSlot
from pyteal.ast.seq import Seq

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class MultiValue(LeafExpr):
    """Represents an operation that returns more than one value"""

    def __init__(
        self,
        op: Op,
        types: List[TealType],
        *,
        immediate_args: List[Union[int, str]] = None,
        args: List[Expr] = None,
        compile_check: Callable[["CompileOptions"], None] = lambda _: None,
    ):
        """Create a new MultiValue.

        Args:
            op: The operation that returns values.
            types: The types of the returned values.
            immediate_args (optional): Immediate arguments for the op. Defaults to None.
            args (optional): Stack arguments for the op. Defaults to None.
        """
        super().__init__()
        self.op = op
        self.types = types
        self.immediate_args = immediate_args if immediate_args is not None else []
        self.args = args if args is not None else []
        self.compile_check = compile_check

        self.output_slots = [ScratchSlot() for _ in self.types]

    def outputReducer(self, reducer: Callable[..., Expr]) -> Expr:
        input = [slot.load(self.types[i]) for i, slot in enumerate(self.output_slots)]
        return Seq(self, reducer(*input))

    def __str__(self):
        ret_str = "(({}".format(self.op)
        for a in self.immediate_args:
            ret_str += " " + a.__str__()

        for a in self.args:
            ret_str += " " + a.__str__()
        ret_str += ") "

        ret_str += " ".join([slot.store().__str__() for slot in self.output_slots])
        ret_str += ")"

        return ret_str

    def __teal__(self, options: "CompileOptions"):
        self.compile_check(options)

        tealOp = TealOp(self, self.op, *self.immediate_args)
        callStart, callEnd = TealBlock.FromOp(options, tealOp, *self.args)

        curEnd = callEnd
        # the list is reversed in order to preserve the ordering of the opcode's returned
        # values. ie the output to stack [A, B, C] should correspond to C->output_slots[2]
        # B->output_slots[1], and A->output_slots[0].
        for slot in reversed(self.output_slots):
            store = slot.store()
            storeStart, storeEnd = store.__teal__(options)
            curEnd.setNextBlock(storeStart)
            curEnd = storeEnd

        return callStart, curEnd

    def type_of(self):
        return TealType.none


MultiValue.__module__ = "pyteal"



from dataclasses import dataclass
from docstring_parser import parse as parse_docstring
from inspect import isclass, Parameter, signature, get_annotations
from types import MappingProxyType, NoneType
from typing import Any, Callable, Final, Optional, TYPE_CHECKING, cast
import algosdk.abi as sdk_abi

from pyteal.ast import abi
from pyteal.ast.expr import Expr
from pyteal.ast.seq import Seq
from pyteal.ast.scratchvar import DynamicScratchVar, ScratchVar
from pyteal.errors import TealInputError, verifyProgramVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.types import TealType

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class SubroutineDefinition:
    """
    Class that leverages TEAL's `callsub` and `retsub` opcode-pair for subroutines
    """

    nextSubroutineId = 0

    def __init__(
        self,
        implementation: Callable[..., Expr],
        return_type: TealType,
        name_str: Optional[str] = None,
        has_abi_output: bool = False,
    ) -> None:
        """
        Args:
            implementation: The python function defining the subroutine
            return_type: the TealType to be returned by the subroutine
            name_str (optional): the name that is used to identify the subroutine.
                If omitted, the name defaults to the implementation's __name__ attribute
            has_abi_output (optional): the boolean that tells if ABI output kwarg for subroutine is used.
        """
        super().__init__()
        self.id = SubroutineDefinition.nextSubroutineId
        SubroutineDefinition.nextSubroutineId += 1

        self.return_type = return_type
        self.declaration: Optional["SubroutineDeclaration"] = None

        self.implementation: Callable = implementation
        self.has_abi_output: bool = has_abi_output

        self.implementation_params: MappingProxyType[str, Parameter]
        self.annotations: dict[str, type]
        self.expected_arg_types: list[type[Expr] | type[ScratchVar] | abi.TypeSpec]
        self.by_ref_args: set[str]
        self.abi_args: dict[str, abi.TypeSpec]
        self.output_kwarg: dict[str, abi.TypeSpec]

        (
            self.implementation_params,
            self.annotations,
            self.expected_arg_types,
            self.by_ref_args,
            self.abi_args,
            self.output_kwarg,
        ) = self._validate()

        self.__name: str = name_str if name_str else self.implementation.__name__

    def _validate(
        self, input_types: list[TealType | None] = None
    ) -> tuple[
        MappingProxyType[str, Parameter],
        dict[str, type],
        list[type[Expr] | type[ScratchVar] | abi.TypeSpec],
        set[str],
        dict[str, abi.TypeSpec],
        dict[str, abi.TypeSpec],
    ]:
        """Validate the full function signature and annotations for subroutine definition.

        NOTE: `self.implementation` should be set before calling `_validate()`

        This function iterates through `sig.parameters.items()`, and checks each of subroutine arguments.
        On each of the subroutine arguments, the following checks are performed:
        - If argument is not POSITION_ONLY or not POSITIONAL_OR_KEYWORD, error
        - If argument has default value, error

        After the previous checks, the function signature is correct in structure,
        but we still need to check the argument types are matching requirements
        (i.e., in {Expr, ScratchVar, inheritances of abi.BaseType}).

        Finally, this function outputs:
        - `implementation_params` - ordered map from parameter name to inspect.Parameter
        - `annotations` - map from parameter name to annotation (if available)
        - `expected_arg_types` - an array of elements of Type[Expr], Type[ScratchVar] or abi.TypeSpec instances
          It helps type-checking on SubroutineCall from `invoke` method.
        - `by_ref_args` - a set of argument names, which are type annotated by ScratchVar.
        We put the scratch slot id on the stack, rather than the value itself.
        - `abi_args` - a set of argument names, which are type annotated by ABI types.
        We load the ABI scratch space stored value to stack, and store them later in subroutine's local ABI values.
        - `abi_output_kwarg` - a possibly empty dict which when non-empty contains exactly one key
            `ABIReturnSubroutine.OUTPUT_ARG_NAME` with a value that gives abi-tye information about the output

        Args:
            input_types (optional): for testing purposes - expected `TealType`s of each parameter

        Returns:
            impl_params: a map from python function implementation's argument name, to argument's parameter.
            annotations: a dict whose keys are names of type-annotated arguments,
                and values are appearing type-annotations.
            arg_types: a list of argument type inferred from python function implementation,
                containing [type[Expr]| type[ScratchVar] | abi.TypeSpec].
            by_ref_args: a list of argument names that are passed in Subroutine with by-reference mechanism.
            abi_args: a dict whose keys are names of ABI arguments, and values are their ABI type-specs.
            abi_output_kwarg (might be empty): a dict whose key is the name of ABI output keyword argument,
                and the value is the corresponding ABI type-spec.
                NOTE: this dict might be empty, when we are defining a normal subroutine,
                    but it has at most one element when we define an ABI-returning subroutine.
        """

        if not callable(self.implementation):
            raise TealInputError("Input to SubroutineDefinition is not callable")

        impl_params: MappingProxyType[str, Parameter] = signature(
            self.implementation
        ).parameters
        annotations: dict[str, type] = get_annotations(self.implementation)
        arg_types: list[type[Expr] | type[ScratchVar] | abi.TypeSpec] = []
        by_ref_args: set[str] = set()
        abi_args: dict[str, abi.TypeSpec] = {}
        abi_output_kwarg: dict[str, abi.TypeSpec] = {}

        if "return" in annotations and annotations["return"] is not Expr:
            raise TealInputError(
                f"Function has return of disallowed type {annotations['return']}. Only Expr is allowed"
            )

        for name, param in impl_params.items():
            if param.kind not in (
                Parameter.POSITIONAL_ONLY,
                Parameter.POSITIONAL_OR_KEYWORD,
            ) and not (
                param.kind is Parameter.KEYWORD_ONLY
                and self.has_abi_output
                and name == ABIReturnSubroutine.OUTPUT_ARG_NAME
            ):
                raise TealInputError(
                    f"Function has a parameter type that is not allowed in a subroutine: parameter {name} with type {param.kind}"
                )

            if param.default != Parameter.empty:
                raise TealInputError(
                    f"Function has a parameter with a default value, which is not allowed in a subroutine: {name}"
                )

            expected_arg_type = self._validate_annotation(annotations, name)

            if param.kind is Parameter.KEYWORD_ONLY:
                # this case is only entered when
                # - `self.has_abi_output is True`
                # - `name == ABIReturnSubroutine.OUTPUT_ARG_NAME`
                if not isinstance(expected_arg_type, abi.TypeSpec):
                    raise TealInputError(
                        f"Function keyword parameter {name} has type {expected_arg_type}"
                    )
                abi_output_kwarg[name] = expected_arg_type
                continue

            arg_types.append(expected_arg_type)
            if expected_arg_type is ScratchVar:
                by_ref_args.add(name)
            if isinstance(expected_arg_type, abi.TypeSpec):
                abi_args[name] = expected_arg_type

        if input_types is not None:
            input_arg_count = len(impl_params) - len(abi_output_kwarg)
            if len(input_types) != input_arg_count:
                raise TealInputError(
                    f"Provided number of input_types ({len(input_types)}) "
                    f"does not match detected number of input parameters ({input_arg_count})"
                )
            for in_type, name in zip(input_types, impl_params):
                if not isinstance(in_type, (TealType, NoneType)):
                    raise TealInputError(
                        f"Function has input type {in_type} for parameter {name} which is not a TealType"
                    )
                if in_type is None and name not in abi_args:
                    raise TealInputError(
                        f"input_type for {name} is unspecified i.e. None "
                        f"but this is only allowed for ABI arguments"
                    )

        return (
            impl_params,
            annotations,
            arg_types,
            by_ref_args,
            abi_args,
            abi_output_kwarg,
        )

    @staticmethod
    def _is_abi_annotation(obj: Any) -> bool:
        try:
            abi.type_spec_from_annotation(obj)
            return True
        except TypeError:
            return False

    @staticmethod
    def _validate_annotation(
        user_defined_annotations: dict[str, Any], parameter_name: str
    ) -> type[Expr] | type[ScratchVar] | abi.TypeSpec:
        ptype = user_defined_annotations.get(parameter_name, None)
        if ptype is None:
            # Without a type annotation, `SubroutineDefinition` presumes an implicit `Expr` declaration
            # rather than these alternatives:
            # * Throw error requiring type annotation.
            # * Defer parameter type checks until arguments provided during invocation.
            #
            # * Rationale:
            #   * Provide an upfront, best-effort type check before invocation.
            #   * Preserve backwards compatibility with TEAL programs written
            #     when `Expr` is the only supported annotation type.
            # * `invoke` type checks provided arguments against parameter types to catch mismatches.
            return Expr
        if ptype in (Expr, ScratchVar):
            return ptype
        if SubroutineDefinition._is_abi_annotation(ptype):
            return abi.type_spec_from_annotation(ptype)
        if not isclass(ptype):
            raise TealInputError(
                f"Function has parameter {parameter_name} of declared type {ptype} which is not a class"
            )
        raise TealInputError(
            f"Function has parameter {parameter_name} of disallowed type {ptype}. "
            f"Only the types {(Expr, ScratchVar, 'ABI')} are allowed"
        )

    def get_declaration(self) -> "SubroutineDeclaration":
        if self.declaration is None:
            # lazy evaluate subroutine
            self.declaration = evaluate_subroutine(self)
        return self.declaration

    def name(self) -> str:
        return self.__name

    def argument_count(self) -> int:
        return len(self.arguments())

    def arguments(self) -> list[str]:
        syntax_args = list(self.implementation_params.keys())
        syntax_args = [
            arg_name for arg_name in syntax_args if arg_name not in self.output_kwarg
        ]
        return syntax_args

    def invoke(
        self,
        args: list[Expr | ScratchVar | abi.BaseType],
    ) -> "SubroutineCall":
        from pyteal.ast.abi.util import type_spec_is_assignable_to

        if len(args) != self.argument_count():
            raise TealInputError(
                f"Incorrect number of arguments for subroutine call. "
                f"Expected {self.arguments()} arguments, got {len(args)} arguments"
            )

        for i, arg in enumerate(args):
            arg_type = self.expected_arg_types[i]
            if arg_type is Expr and not isinstance(arg, Expr):
                raise TealInputError(
                    f"supplied argument {arg} at index {i} had type {type(arg)} but was expecting type {arg_type}"
                )
            elif arg_type is ScratchVar and not isinstance(arg, ScratchVar):
                raise TealInputError(
                    f"supplied argument {arg} at index {i} had type {type(arg)} but was expecting type {arg_type}"
                )
            elif isinstance(arg_type, abi.TypeSpec):
                if not isinstance(arg, abi.BaseType):
                    raise TealInputError(
                        f"supplied argument at index {i} should be an ABI type but got {arg}"
                    )

                if not type_spec_is_assignable_to(arg.type_spec(), arg_type):
                    raise TealInputError(
                        f"supplied argument {arg} at index {i} "
                        f"should have ABI typespec {arg_type} but got {arg.type_spec()}"
                    )

        return SubroutineCall(
            self, args, output_kwarg=OutputKwArgInfo.from_dict(self.output_kwarg)
        )

    def __str__(self):
        return f"subroutine#{self.id}"

    def __eq__(self, other):
        if isinstance(other, SubroutineDefinition):
            return self.id == other.id and self.implementation == other.implementation
        return False

    def __hash__(self):
        return hash(self.id)


SubroutineDefinition.__module__ = "pyteal"


class SubroutineDeclaration(Expr):
    def __init__(
        self,
        subroutine: SubroutineDefinition,
        body: Expr,
        deferred_expr: Optional[Expr] = None,
    ) -> None:
        super().__init__()
        self.subroutine = subroutine
        self.body = body
        self.deferred_expr = deferred_expr

    def __teal__(self, options: "CompileOptions"):
        return self.body.__teal__(options)

    def __str__(self):
        return f'(SubroutineDeclaration "{self.subroutine.name()}" {self.body})'

    def type_of(self):
        return self.body.type_of()

    def has_return(self):
        return self.body.has_return()


SubroutineDeclaration.__module__ = "pyteal"


@dataclass
class OutputKwArgInfo:
    name: str
    abi_type: abi.TypeSpec

    @staticmethod
    def from_dict(kwarg_info: dict[str, abi.TypeSpec]) -> Optional["OutputKwArgInfo"]:
        match list(kwarg_info.keys()):
            case []:
                return None
            case [k]:
                return OutputKwArgInfo(k, kwarg_info[k])
            case _:
                raise TealInputError(
                    f"illegal conversion kwarg_info length {len(kwarg_info)}."
                )


OutputKwArgInfo.__module__ = "pyteal"


class SubroutineCall(Expr):
    def __init__(
        self,
        subroutine: SubroutineDefinition,
        args: list[Expr | ScratchVar | abi.BaseType],
        *,
        output_kwarg: OutputKwArgInfo = None,
    ) -> None:
        super().__init__()
        self.subroutine = subroutine
        self.args = args
        self.output_kwarg = output_kwarg

        for i, arg in enumerate(args):
            if isinstance(arg, Expr):
                arg_type = arg.type_of()
            elif isinstance(arg, ScratchVar):
                arg_type = arg.type
            elif isinstance(arg, abi.BaseType):
                arg_type = cast(abi.BaseType, arg).stored_value.type
            else:
                raise TealInputError(
                    f"Subroutine argument {arg} at index {i} was of unexpected Python type {type(arg)}"
                )

            if arg_type == TealType.none:
                raise TealInputError(
                    f"Subroutine argument {arg} at index {i} evaluates to TealType.none"
                )

    def __teal__(self, options: "CompileOptions"):
        """
        Generate the subroutine's start and end teal blocks.
        The subroutine's arguments are pushed on the stack to be picked up into local scratch variables.
        There are 4 cases to consider for the pushed arg expression:

        1. (by-value) In the case of typical arguments of type Expr, the expression ITSELF is evaluated for the stack
            and will be stored in a local ScratchVar for subroutine evaluation

        2. (by-reference) In the case of a by-reference argument of type ScratchVar, its SLOT INDEX is put on the stack
            and will be stored in a local DynamicScratchVar for subroutine evaluation

        3. (ABI, or a special case in by-value) In this case, the storage of an ABI value are loaded
            to the stack and will be stored in a local ABI value for subroutine evaluation

        4. (ABI output keyword argument) In this case, we do not place ABI values (encoding) on the stack.
            This is an *output-only* argument: in `evaluate_subroutine` an ABI typed instance for subroutine evaluation
            will be generated, and gets in to construct the subroutine implementation.
        """
        verifyProgramVersion(
            Op.callsub.min_version,
            options.version,
            "Program version too low to use SubroutineCall expression",
        )

        def handle_arg(arg: Expr | ScratchVar | abi.BaseType) -> Expr:
            if isinstance(arg, ScratchVar):
                return arg.index()
            elif isinstance(arg, Expr):
                return arg
            elif isinstance(arg, abi.BaseType):
                return arg.stored_value.load()
            else:
                raise TealInputError(
                    f"cannot handle current arg: {arg} to put it on stack"
                )

        op = TealOp(self, Op.callsub, self.subroutine)
        return TealBlock.FromOp(options, op, *[handle_arg(x) for x in self.args])

    def __str__(self):
        arg_str_list = list(map(str, self.args))
        if self.output_kwarg:
            arg_str_list.append(
                f"{self.output_kwarg.name}={str(self.output_kwarg.abi_type)}"
            )
        return f'(SubroutineCall {self.subroutine.name()} ({" ".join(arg_str_list)}))'

    def type_of(self):
        return self.subroutine.return_type

    def has_return(self):
        return False


SubroutineCall.__module__ = "pyteal"


class SubroutineFnWrapper:
    def __init__(
        self,
        fn_implementation: Callable[..., Expr],
        return_type: TealType,
        name: Optional[str] = None,
    ) -> None:
        self.subroutine = SubroutineDefinition(
            fn_implementation,
            return_type=return_type,
            name_str=name,
        )

    def __call__(self, *args: Expr | ScratchVar | abi.BaseType, **kwargs: Any) -> Expr:
        if len(kwargs) != 0:
            raise TealInputError(
                f"Subroutine cannot be called with keyword arguments. "
                f"Received keyword arguments: {','.join(kwargs.keys())}"
            )
        return self.subroutine.invoke(list(args))

    def name(self) -> str:
        return self.subroutine.name()

    def type_of(self):
        return self.subroutine.get_declaration().type_of()

    def has_return(self):
        return self.subroutine.get_declaration().has_return()


SubroutineFnWrapper.__module__ = "pyteal"


class ABIReturnSubroutine:
    """Used to create a PyTeal Subroutine (returning an ABI value) from a python function.  It's primarily intended to define ARC-4 Application entry points though it can also be used more generally.

    *Disclaimer*:  ABIReturnSubroutine is still taking shape and is subject to backwards incompatible changes.

    * For ARC-4 Application entry point definition, feel encouraged to use ABIReturnSubroutine.  Expect a best-effort attempt to minimize backwards incompatible changes along with a migration path.
    * For general purpose subroutine definition usage, use at your own risk.  Based on feedback, the API and usage patterns will change more freely and with less effort to provide migration paths.

    This class is meant to be used as a function decorator. For example:

        .. code-block:: python

            @ABIReturnSubroutine
            def abi_sum(toSum: abi.DynamicArray[abi.Uint64], *, output: abi.Uint64) -> Expr:
                i = ScratchVar(TealType.uint64)
                valueAtIndex = abi.Uint64()
                return Seq(
                    output.set(0),
                    For(i.store(Int(0)), i.load() < toSum.length(), i.store(i.load() + Int(1))).Do(
                        Seq(
                            toSum[i.load()].store_into(valueAtIndex),
                            output.set(output.get() + valueAtIndex.get()),
                        )
                    ),
                )

            program = Seq(
                (to_sum_arr := abi.make(abi.DynamicArray[abi.Uint64])).decode(
                    Txn.application_args[1]
                ),
                (res := abi.Uint64()).set(abi_sum(to_sum_arr)),
                abi.MethodReturn(res),
                Int(1),
            )
    """

    OUTPUT_ARG_NAME: Final[str] = "output"

    def __init__(
        self,
        fn_implementation: Callable[..., Expr],
        /,
        *,
        overriding_name: Optional[str] = None,
    ) -> None:
        self.output_kwarg_info: Optional[OutputKwArgInfo] = self._get_output_kwarg_info(
            fn_implementation
        )
        self.subroutine = SubroutineDefinition(
            fn_implementation,
            return_type=TealType.none,
            name_str=overriding_name,
            has_abi_output=self.output_kwarg_info is not None,
        )

    @staticmethod
    def name_override(name: str) -> Callable[..., "ABIReturnSubroutine"]:
        def wrapper(fn_impl: Callable[..., Expr]) -> ABIReturnSubroutine:
            return ABIReturnSubroutine(fn_impl, overriding_name=name)

        return wrapper

    @classmethod
    def _get_output_kwarg_info(
        cls, fn_implementation: Callable[..., Expr]
    ) -> Optional[OutputKwArgInfo]:
        if not callable(fn_implementation):
            raise TealInputError("Input to ABIReturnSubroutine is not callable")
        sig = signature(fn_implementation)
        fn_annotations = get_annotations(fn_implementation)

        potential_abi_arg_names = [
            k for k, v in sig.parameters.items() if v.kind == Parameter.KEYWORD_ONLY
        ]

        match potential_abi_arg_names:
            case []:
                return None
            case [name]:
                if name != cls.OUTPUT_ARG_NAME:
                    raise TealInputError(
                        f"ABI return subroutine output-kwarg name must be `output` at this moment, "
                        f"while {name} is the keyword."
                    )
                annotation = fn_annotations.get(name, None)
                if annotation is None:
                    raise TealInputError(
                        f"ABI return subroutine output-kwarg {name} must specify ABI type"
                    )
                type_spec = abi.type_spec_from_annotation(annotation)
                return OutputKwArgInfo(name, type_spec)
            case _:
                raise TealInputError(
                    f"multiple output arguments ({len(potential_abi_arg_names)}) "
                    f"with type annotations {potential_abi_arg_names}"
                )

    def __call__(
        self, *args: Expr | ScratchVar | abi.BaseType, **kwargs
    ) -> abi.ReturnedValue | Expr:
        if len(kwargs) != 0:
            raise TealInputError(
                f"Subroutine cannot be called with keyword arguments. "
                f"Received keyword arguments: {', '.join(kwargs.keys())}"
            )

        invoked = self.subroutine.invoke(list(args))
        if self.output_kwarg_info is None:
            if invoked.type_of() != TealType.none:
                raise TealInputError(
                    "ABI subroutine with void type should be evaluated to TealType.none"
                )
            return invoked

        self.subroutine.get_declaration()

        return abi.ReturnedValue(
            self.output_kwarg_info.abi_type,
            invoked,
        )

    def name(self) -> str:
        return self.subroutine.name()

    def method_signature(self, overriding_name: str = None) -> str:
        if not self.is_abi_routable():
            raise TealInputError(
                "Only registrable methods may return a method signature"
            )

        ret_type = self.type_of()
        if isinstance(ret_type, abi.TypeSpec) and abi.contains_type_spec(
            ret_type, abi.TransactionTypeSpecs + abi.ReferenceTypeSpecs
        ):
            raise TealInputError(
                f"Reference and Transaction types may not be used as return values, got {ret_type}"
            )

        args = [str(v) for v in self.subroutine.abi_args.values()]
        if overriding_name is None:
            overriding_name = self.name()
        return f"{overriding_name}({','.join(args)}){self.type_of()}"

    def method_spec(self) -> sdk_abi.Method:
        desc: str = ""
        arg_descs: dict[str, str] = {}
        return_desc: str = ""
        args: list[dict[str, str]] = []

        if self.subroutine.implementation.__doc__:
            docstring = parse_docstring(self.subroutine.implementation.__doc__)

            # Combine short and long descriptions with newline
            method_descriptions: list[str] = []
            if docstring.short_description:
                method_descriptions.append(docstring.short_description)
            if docstring.long_description:
                method_descriptions.append(docstring.long_description)

            method_desc = "\n\n".join(method_descriptions)

            # Turn double new line into single, replacing single newline with space
            desc = "\n".join(
                [
                    i.replace("\n", " ").strip()
                    for i in method_desc.split("\n\n")
                    if i.strip()
                ]
            )

            # Get the descriptions for any documented arguments
            arg_descs = {
                arg.arg_name: arg.description.replace("\n", " ").strip()
                for arg in docstring.params
                if arg.arg_name != self.OUTPUT_ARG_NAME and arg.description is not None
            }

            # Get the special return description
            return_desc = (
                ""
                if not docstring.returns or not docstring.returns.description
                else docstring.returns.description.replace("\n", " ").strip()
            )

        # Generate the ABI method object given the subroutine args
        # Add in description if one is set
        for name, val in self.subroutine.annotations.items():
            # Skip annotations for `return` and `output` in the args list
            if name in ["return", self.OUTPUT_ARG_NAME]:
                continue

            arg_obj = {
                "type": str(abi.type_spec_from_annotation(val)),
                "name": name,
            }

            if name in arg_descs:
                arg_obj["desc"] = arg_descs[name]

            args.append(arg_obj)

        # Create the return obj for the method, adding description if set
        return_obj = {"type": str(self.type_of())}
        if return_desc and return_obj["type"] != "void":
            return_obj["desc"] = return_desc

        # Create the method spec, adding description if set
        spec = {"name": self.name(), "args": args, "returns": return_obj}
        if desc:
            spec["desc"] = desc

        return sdk_abi.Method.undictify(spec)

    def type_of(self) -> str | abi.TypeSpec:
        return (
            "void"
            if self.output_kwarg_info is None
            else self.output_kwarg_info.abi_type
        )

    def is_abi_routable(self) -> bool:
        return len(self.subroutine.abi_args) == self.subroutine.argument_count()


ABIReturnSubroutine.__module__ = "pyteal"


class Subroutine:
    """Used to create a PyTeal subroutine from a Python function.

    This class is meant to be used as a function decorator. For example:

        .. code-block:: python

            @Subroutine(TealType.uint64)
            def mySubroutine(a: Expr, b: Expr) -> Expr:
                return a + b

            program = Seq([
                App.globalPut(Bytes("key"), mySubroutine(Int(1), Int(2))),
                Approve(),
            ])
    """

    def __init__(self, return_type: TealType, name: Optional[str] = None) -> None:
        """Define a new subroutine with the given return type.

        Args:
            return_type: The type that the return value of this subroutine must conform to.
                TealType.none indicates that this subroutine does not return any value.
        """
        self.return_type = return_type
        self.name = name

    def __call__(self, fn_implementation: Callable[..., Expr]) -> SubroutineFnWrapper:
        return SubroutineFnWrapper(
            fn_implementation=fn_implementation,
            return_type=self.return_type,
            name=self.name,
        )


Subroutine.__module__ = "pyteal"


def evaluate_subroutine(subroutine: SubroutineDefinition) -> SubroutineDeclaration:
    """
    Puts together the data necessary to define the code for a subroutine.
    "evaluate" is used here to connote evaluating the PyTEAL AST into a SubroutineDeclaration,
    but not actually placing it at call locations. The trickiest part here is managing the subroutine's arguments.
    The arguments are needed for two different code-paths, and there are 2 different argument types to consider
    for each of the code-paths:

    2 Argument Usages / Code-Paths
    - -------- ------   ----------
    Usage (A) for run-time: "argumentVars" --reverse--> "body_ops"
        These are "store" expressions that pick up parameters that have been pre-placed on the stack prior to subroutine invocation.
        The argumentVars are stored into local scratch space to be used by the TEAL subroutine.

    Usage (B) for compile-time: "loadedArgs"
        These are expressions supplied to the user-defined PyTEAL function.
        The loadedArgs are invoked to by the subroutine to create a self-contained AST which will translate into a TEAL subroutine.

    In both usage cases, we need to handle

    2 Argument Types
    - -------- -----
    Type 1 (by-value): these have python type Expr
    Type 2 (by-reference): these have python type ScratchVar
    Type 3 (ABI): these are ABI typed variables with scratch space storage, and still pass by value
    Type 4 (ABI-output-arg): ABI typed variables with scratch space, a new ABI instance is generated inside function body,
        not one of the cases in the previous three options

    Usage (A) "argumentVars" - Storing pre-placed stack variables into local scratch space:
        Type 1. (by-value) use ScratchVar.store() to pick the actual value into a local scratch space
        Type 2. (by-reference) ALSO use ScratchVar.store() to pick up from the stack
            NOTE: SubroutineCall.__teal__() has placed the _SLOT INDEX_ on the stack so this is stored into the local scratch space
        Type 3. (ABI) abi_value.stored_value.store() to pick from the stack
        Type 4. (ABI-output-arg) it is not really used here, since it is only generated internal of the subroutine

    Usage (B) "loadedArgs" - Passing through to an invoked PyTEAL subroutine AST:
        Type 1. (by-value) use ScratchVar.load() to have an Expr that can be compiled in python by the PyTEAL subroutine
        Type 2. (by-reference) use a DynamicScratchVar as the user will have written the PyTEAL in a way that satisfies
            the ScratchVar API. I.e., the user will write `x.load()` and `x.store(val)` as opposed to just `x`.
        Type 3. (ABI) use abi_value itself after storing stack value into scratch space.
        Type 4. (ABI-output-arg) generates a new instance of the ABI value,
            and appends a return expression of stored value of the ABI keyword value.
    """

    def var_n_loaded(
        param: str,
    ) -> tuple[ScratchVar, ScratchVar | abi.BaseType | Expr]:
        loaded_var: ScratchVar | abi.BaseType | Expr
        argument_var: ScratchVar

        if param in subroutine.by_ref_args:
            argument_var = DynamicScratchVar(TealType.anytype)
            loaded_var = argument_var
        elif param in subroutine.abi_args:
            internal_abi_var = subroutine.abi_args[param].new_instance()
            argument_var = internal_abi_var.stored_value
            loaded_var = internal_abi_var
        else:
            argument_var = ScratchVar(TealType.anytype)
            loaded_var = argument_var.load()

        return argument_var, loaded_var

    if len(subroutine.output_kwarg) > 1:
        raise TealInputError(
            f"ABI keyword argument num: {len(subroutine.output_kwarg)}. "
            f"Exceeding abi output keyword argument max number 1."
        )

    args = subroutine.arguments()

    arg_vars: list[ScratchVar] = []
    loaded_args: list[ScratchVar | Expr | abi.BaseType] = []
    for arg in args:
        arg_var, loaded_arg = var_n_loaded(arg)
        arg_vars.append(arg_var)
        loaded_args.append(loaded_arg)

    abi_output_kwargs: dict[str, abi.BaseType] = {}
    output_kwarg_info = OutputKwArgInfo.from_dict(subroutine.output_kwarg)
    output_carrying_abi: Optional[abi.BaseType] = None

    if output_kwarg_info:
        output_carrying_abi = output_kwarg_info.abi_type.new_instance()
        abi_output_kwargs[output_kwarg_info.name] = output_carrying_abi

    # Arg usage "B" supplied to build an AST from the user-defined PyTEAL function:
    subroutine_body = subroutine.implementation(*loaded_args, **abi_output_kwargs)

    if not isinstance(subroutine_body, Expr):
        raise TealInputError(
            f"Subroutine function does not return a PyTeal expression. Got type {type(subroutine_body)}."
        )

    deferred_expr: Optional[Expr] = None

    # if there is an output keyword argument for ABI, place the storing on the stack
    if output_carrying_abi:
        if subroutine_body.type_of() != TealType.none:
            raise TealInputError(
                f"ABI returning subroutine definition should evaluate to TealType.none, "
                f"while evaluate to {subroutine_body.type_of()}."
            )
        deferred_expr = output_carrying_abi.stored_value.load()

    # Arg usage "A" to be pick up and store in scratch parameters that have been placed on the stack
    # need to reverse order of argumentVars because the last argument will be on top of the stack
    body_ops = [var.slot.store() for var in arg_vars[::-1]]
    body_ops.append(subroutine_body)

    sd = SubroutineDeclaration(subroutine, Seq(body_ops), deferred_expr)
    sd.trace = subroutine_body.trace
    return sd



import pytest

import pyteal as pt

avm2Options = pt.CompileOptions(version=2)
avm4Options = pt.CompileOptions(version=4)
avm5Options = pt.CompileOptions(version=5)


def test_arg_static():
    for i in range(256):
        expr = pt.Arg(i)
        assert expr.type_of() == pt.TealType.bytes
        assert not expr.has_return()

        expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.arg, i)])

        actual, _ = expr.__teal__(avm2Options)
        assert actual == expected


def test_arg_dynamic():
    i = pt.Int(7)
    expr = pt.Arg(i)
    assert expr.type_of() == pt.TealType.bytes
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [pt.TealOp(i, pt.Op.int, 7), pt.TealOp(expr, pt.Op.args)]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm4Options)


def test_arg_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Arg(pt.Bytes("k"))

    with pytest.raises(pt.TealInputError):
        pt.Arg(-1)

    with pytest.raises(pt.TealInputError):
        pt.Arg(256)



from typing import Optional, List

from pyteal.ir.tealop import TealOp
from pyteal.ir.tealblock import TealBlock


class TealSimpleBlock(TealBlock):
    """Represents a basic block of TealComponents in a graph that does not contain a branch condition."""

    def __init__(self, ops: List[TealOp]) -> None:
        super().__init__(ops)
        self.nextBlock: Optional[TealBlock] = None
        self.visited = False

    def setNextBlock(self, block: TealBlock) -> None:
        """Set the block that follows this one."""
        self.nextBlock = block

    def getOutgoing(self) -> List[TealBlock]:
        if self.nextBlock is None:
            return []
        return [self.nextBlock]

    def replaceOutgoing(self, oldBlock: TealBlock, newBlock: TealBlock) -> None:
        if self.nextBlock is oldBlock:
            self.nextBlock = newBlock

    def __repr__(self) -> str:
        # check for loop
        if self.visited:
            return "TealSimpleBlock({}, next={})".format(
                repr(self.ops),
                "",
            )
        self.visited = True

        s = "TealSimpleBlock({}, next={})".format(
            repr(self.ops),
            repr(self.nextBlock),
        )

        self.visited = False
        return s

    def __eq__(self, other: object) -> bool:
        # check for loop
        if self.visited:
            return True
        if type(other) is not TealSimpleBlock:
            return False
        self.visited = True
        equal = self.ops == other.ops and self.nextBlock == other.nextBlock
        self.visited = False
        return equal


TealSimpleBlock.__module__ = "pyteal"



from pyteal.ir.ops import Op, Mode

from pyteal.ir.tealcomponent import TealComponent
from pyteal.ir.tealop import TealOp
from pyteal.ir.teallabel import TealLabel
from pyteal.ir.tealblock import TealBlock
from pyteal.ir.tealsimpleblock import TealSimpleBlock
from pyteal.ir.tealconditionalblock import TealConditionalBlock

from pyteal.ir.labelref import LabelReference

__all__ = [
    "Op",
    "Mode",
    "TealComponent",
    "TealOp",
    "TealLabel",
    "TealBlock",
    "TealSimpleBlock",
    "TealConditionalBlock",
    "LabelReference",
]



from typing import cast, Union, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.errors import TealInputError, verifyProgramVersion
from pyteal.config import MAX_GROUP_SIZE, NUM_SLOTS
from pyteal.ast.expr import Expr
from pyteal.ast.leafexpr import LeafExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class ImportScratchValue(LeafExpr):
    """An expression to load a scratch value created by another transaction in the current group"""

    def __init__(self, txnIndex: Union[int, Expr], slotId: Union[int, Expr]) -> None:
        """Create an expression to load a scratch space slot from a transaction in the current group.

        Requires program version 4 or higher. This operation is only permitted in application mode.

        Args:
            txnIndex: The index of the transaction from which the created ID should be obtained.
                This index may be a Python int, or it may be a PyTeal expression that evaluates at
                runtime. If it's an expression, it must evaluate to a uint64. In all cases, the index
                must be less than the index of the current transaction.
            slotId: The index of the scratch slot that should be loaded.
                This index may be a Python int, or it may be a PyTeal expression that evaluates at
                runtime. If it's an expression, it must evaluate to a uint64. In all cases, the index
                must be in the range [0, 256).
        """
        super().__init__()
        if type(txnIndex) is int:
            if txnIndex < 0 or txnIndex >= MAX_GROUP_SIZE:
                raise TealInputError(
                    "Invalid transaction index {}, shoud be in [0, {})".format(
                        txnIndex, MAX_GROUP_SIZE
                    )
                )
            if type(slotId) is not int:
                raise TealInputError(
                    "Invalid gload syntax with immediate transaction index {} and stack slot ID {}".format(
                        txnIndex, slotId
                    )
                )
        else:
            require_type(cast(Expr, txnIndex), TealType.uint64)

        if type(slotId) is int:
            if slotId < 0 or slotId >= NUM_SLOTS:
                raise TealInputError(
                    "Invalid slot ID {}, shoud be in [0, {})".format(slotId, NUM_SLOTS)
                )
        else:
            require_type(cast(Expr, slotId), TealType.uint64)

        self.txnIndex = txnIndex
        self.slotId = slotId

    def __str__(self) -> str:
        return "(Gload {} {})".format(self.txnIndex, self.slotId)

    def __teal__(self, options: "CompileOptions"):
        def local_version_check(opcode: TealOp):
            verifyProgramVersion(
                opcode.op.min_version,
                options.version,
                "Program version too low to use {} experssion".format(opcode.op.name),
            )

        # For txnIndex and slotId, there are only three scenario as following
        #     immediate    immediate
        #     stack        immediate
        #     stack        stack
        # the last one is not allowed
        # --> immediate    stack
        # which is eliminated in __init__
        if type(self.txnIndex) is int and type(self.slotId) is int:
            op = TealOp(self, Op.gload, self.txnIndex, self.slotId)
            local_version_check(op)
            return TealBlock.FromOp(options, op)
        elif type(self.slotId) is int:
            op = TealOp(self, Op.gloads, self.slotId)
            local_version_check(op)
            return TealBlock.FromOp(options, op, cast(Expr, self.txnIndex))
        else:
            op = TealOp(self, Op.gloadss)
            local_version_check(op)
            return TealBlock.FromOp(
                options, op, cast(Expr, self.txnIndex), cast(Expr, self.slotId)
            )

    def type_of(self):
        return TealType.anytype


ImportScratchValue.__module__ = "pyteal"



from typing import Final

from pyteal.types import TealType, require_type
from pyteal.ir import Op
from pyteal.ast.expr import Expr
from pyteal.ast.maybe import MaybeValue


class AccountParam:
    @classmethod
    def balance(cls, acct: Expr) -> MaybeValue:
        """Get the current balance in microalgos an account.

        Args:
            acct: An index into Txn.accounts that corresponds to the application to check or an address available at runtime.
                May evaluate to uint64 or an address.
        """
        require_type(acct, TealType.anytype)
        return MaybeValue(
            Op.acct_params_get,
            TealType.uint64,
            immediate_args=["AcctBalance"],
            args=[acct],
        )

    @classmethod
    def minBalance(cls, acct: Expr) -> MaybeValue:
        """Get the minimum balance in microalgos for an account.

        Args:
            acct: An index into Txn.accounts that corresponds to the application to check or an address available at runtime.
                May evaluate to uint64 or an address.
        """
        require_type(acct, TealType.anytype)
        return MaybeValue(
            Op.acct_params_get,
            TealType.uint64,
            immediate_args=["AcctMinBalance"],
            args=[acct],
        )

    @classmethod
    def authAddr(cls, acct: Expr) -> MaybeValue:
        """Get the authorizing address for an account. If the account is not rekeyed, the empty addresss is returned.

        Args:
            acct: An index into Txn.accounts that corresponds to the application to check or an address available at runtime.
                May evaluate to uint64 or an address.
        """
        require_type(acct, TealType.anytype)
        return MaybeValue(
            Op.acct_params_get,
            TealType.bytes,
            immediate_args=["AcctAuthAddr"],
            args=[acct],
        )


AccountParam.__module__ = "pyteal"


class AccountParamObject:
    """Represents information about an account"""

    def __init__(self, account: Expr) -> None:
        """Create a new AccountParamObject for the given account.

        Args:
            account: An index into Txn.accounts that corresponds to the application to check or an
                address available at runtime. May evaluate to uint64 or bytes, respectively.
        """
        self._account: Final = account

    def balance(self) -> MaybeValue:
        """Get the current balance in microAlgos for the account"""
        return AccountParam.balance(self._account)

    def min_balance(self) -> MaybeValue:
        """Get the minimum balance in microAlgos for the account."""
        return AccountParam.minBalance(self._account)

    def auth_address(self) -> MaybeValue:
        """Get the authorizing address for the account.

        If the account is not rekeyed, the empty address is returned."""
        return AccountParam.authAddr(self._account)


AccountParamObject.__module__ = "pyteal"
